#ifndef FrameCPP__COMMON__PTR_STRUCT_HH
#define FrameCPP__COMMON__PTR_STRUCT_HH

#include <iostream>
#include <string>

#include "general/types.hh"

#include "framecpp/Common/FrameSpec.hh"
#include "framecpp/Common/PTR_STRUCT_BASE.hh"
#include "framecpp/Common/IOStream.hh"

namespace FrameCPP
{
  namespace Common
  {
    template< class DC, class DI >
    struct PTR_STRUCT
      : PTR_STRUCT_BASE
    {
      typedef DC data_class_type;
      typedef DI data_instance_type;

      typedef PTR_STRUCT_BASE::class_type class_type;
      typedef PTR_STRUCT_BASE::instance_type instance_type;

      data_class_type		data_class;
      data_instance_type	data_instance;

      PTR_STRUCT( );

      PTR_STRUCT( data_class_type ClassId, data_instance_type Instance );

      virtual class_type Class( ) const;

      virtual void Class( class_type Class );

      virtual instance_type Instance( ) const;

      virtual void Instance( instance_type Instance );

      const PTR_STRUCT& operator=( const PTR_STRUCT& RHS );
      static cmn_streamsize_type Bytes( );

      cmn_streamsize_type Bytes( const StreamBase& Stream ) const;

      virtual PTR_STRUCT* Create( IStream& Stream ) const;

      virtual PTR_STRUCT* Clone( ) const;

      virtual const char* const ObjectStructName( ) const;

      virtual void Write( OStream& Stream ) const;


      /// \brief comparison operator
      bool operator==( const PTR_STRUCT& RHS ) const;

      /// \brief comparison operator
      virtual bool operator==( const Common::FrameSpec::Object& Obj ) const;

      bool operator!=( const PTR_STRUCT& RHS ) const;

      static const PTR_STRUCT& NULL_PTR;

    protected:
      /// \brief Down grade an object
      virtual std::auto_ptr< Object >
      demote( INT_2U Target, IStream* Stream ) const;

      /// \brief Upgrade an object
      virtual std::auto_ptr< Object >
      promote( INT_2U Target, Object* Obj, IStream* Stream ) const;

    private:
      using FrameSpec::Object::Create;

      PTR_STRUCT( IStream& Stream );
      PTR_STRUCT( const PTR_STRUCT& Source );

    };

    template< class DC, class DI >
    inline PTR_STRUCT< DC, DI >::
    PTR_STRUCT( )
      : data_class( 0 ),
	data_instance( 0 )
    {
    }

    template< class DC, class DI >
    inline PTR_STRUCT< DC, DI >::
    PTR_STRUCT( data_class_type ClassId, data_instance_type Instance )
      : data_class( ClassId ),
	data_instance( Instance )
    {
    }

    
    template< class DC, class DI >
    inline PTR_STRUCT< DC, DI >::
    PTR_STRUCT( IStream& Stream )
    {
      Stream >> data_class
	     >> data_instance
	;
    }

    template< class DC, class DI >
    inline PTR_STRUCT< DC, DI >::
    PTR_STRUCT( const PTR_STRUCT& Source )
      : data_class( Source.data_class ),
	data_instance( Source.data_instance )
    {
    }

    template< class DC, class DI >
    typename PTR_STRUCT< DC, DI >::class_type PTR_STRUCT< DC, DI >::
    Class( ) const
    {
      return data_class;
    }

    template< class DC, class DI >
    void PTR_STRUCT< DC, DI >::
    Class( class_type Class )
    {
      data_class = Class;
    }

    template< class DC, class DI >
    typename PTR_STRUCT< DC, DI >::instance_type PTR_STRUCT< DC, DI >::
    Instance( ) const
    {
      return data_instance;
    }

    template< class DC, class DI >
    void PTR_STRUCT< DC, DI >::
    Instance( instance_type Instance )
    {
      data_instance = Instance;
    }

    template< class DC, class DI >
    const PTR_STRUCT< DC, DI >& PTR_STRUCT< DC, DI >::
    operator= ( const PTR_STRUCT< DC, DI >& RHS )
    {
      data_class = RHS.data_class;
      data_instance = RHS.data_instance;
      return *this;
    }

    template< class DC, class DI >
    inline cmn_streamsize_type PTR_STRUCT< DC, DI >::
    Bytes( )
    {
      return
	sizeof( data_class_type )
	+ sizeof( data_instance_type )
	;
    }
 
    template< class DC, class DI >
    inline cmn_streamsize_type PTR_STRUCT< DC, DI >::
    Bytes( const StreamBase& Stream ) const
    {
      return Bytes( );
    }

    template< class DC, class DI >
    inline PTR_STRUCT< DC, DI>* PTR_STRUCT< DC, DI >::
    Create( IStream& Stream ) const
    {
      return new PTR_STRUCT( Stream );
    }

    template< class DC, class DI >
    inline PTR_STRUCT< DC, DI>* PTR_STRUCT< DC, DI >::
    Clone( ) const
    {
      return new PTR_STRUCT( *this );
    }

    template< class DC, class DI >
    inline std::auto_ptr< FrameSpec::Object > PTR_STRUCT< DC, DI >::
    demote( INT_2U Target, IStream* Stream ) const
    {
      throw
	Unimplemented( "Object* PTR_STRUCT< DC, DI>* PTR_STRUCT< DC, DI >::Demote( Object* Obj ) const",
		       0, __FILE__, __LINE__ );
    }

    template< class DC, class DI >
    inline std::auto_ptr< FrameSpec::Object > PTR_STRUCT< DC, DI >::
    promote( INT_2U Target, Object* Obj, IStream* Stream ) const
    {
      throw
	Unimplemented( "Object* PTR_STRUCT< DC, DI>* PTR_STRUCT< DC, DI >::Promote( Object* Obj ) const",
		       0, __FILE__, __LINE__ );
    }

    template< class DC, class DI >
    inline const char* const PTR_STRUCT< DC, DI >::
    ObjectStructName( ) const
    {
      static const char* const name = "PTR_STRUCT";
      return name;
    }

    template< class DC, class DI >
    inline void PTR_STRUCT< DC, DI >::
    Write( OStream& Stream ) const
    {
      Stream << data_class
	     << data_instance
	;
    }

    template< class DC, class DI >
    inline bool PTR_STRUCT< DC, DI >::
    operator==( const PTR_STRUCT& RHS ) const
    {
      return ( ( data_class == RHS.data_class )
	       && ( data_instance == RHS.data_instance ) );
    }

    template< class DC, class DI >
    inline bool PTR_STRUCT< DC, DI >::
    operator==( const Common::FrameSpec::Object& Obj ) const
    {
      return ( *this == Obj );
    }

    template< class DC, class DI >
    inline bool PTR_STRUCT< DC, DI >::
    operator!=( const PTR_STRUCT< DC, DI >& RHS ) const
    {
      return ! ( *this == RHS );
    }

    template< class DC, class DI >
    const PTR_STRUCT< DC, DI >& PTR_STRUCT< DC, DI >::NULL_PTR =
      PTR_STRUCT< DC, DI >( 0, 0 );
  } // namespace - Common
} // namespace - FrameCPP
#endif /* FrameCPP__COMMON__PTR_STRUCT_HH */
